﻿#include "Dialog.h"
#include <QPaintEvent>
#include <QKeyEvent>
#include <QPainter>
#include <QTimer>
#include <QLinearGradient>
#include <QTime>

namespace {
    QList<QPoint> targetList = {
        QPoint(4 * 4, 93 * 4),
        QPoint(35 * 4, 93 * 4),
        QPoint(57 * 4, 93 * 4),
        QPoint(78 * 4, 81 * 4),
        QPoint(90 * 4, 93 * 4),
        QPoint(112 * 4, 93 * 4),
        QPoint(133 * 4, 93 * 4),
        QPoint(150 * 4, 81 * 4),
        QPoint(160 * 4, 93 * 4)
    };
    QPoint startPoint = QPoint(21 * 4, 56 * 4);
    qreal vel1 = -20.0;
    qreal vel2_slow = -10.0;
    qreal xspeed = 6;
    qreal radius = 4.0;
    QPoint endPoint = QPoint(177 * 4, 76 * 4);//183
    int dotMax = 60;
    int starMax = 7;

    QRect groundPos = QRect(0, 93 * 4, 720, 93 * 4 + 60);
}

Dialog::Dialog(QWidget *parent)
    : QWidget(parent)
    , sndNz1(new QSound(":/snd/se/nz1.wav", this))
    , sndDoor(new QSound(":/snd/se/door-close2.wav"))
{
    this->setWindowFlags(this->windowFlags() & ~Qt::WindowContextHelpButtonHint);
    QTime currenttime = QTime::currentTime();
    qsrand(static_cast<unsigned int>(currenttime.msec() * 60 + currenttime.second()));
    this->setWindowTitle("WAY BACK");
    this->dots.resize(static_cast<size_t>(::dotMax));
    this->stars.resize(static_cast<size_t>(::starMax));
    this->initializeAll();
    this->imgWayback.load(":/images/pic/wayback.png");
    this->imgPlayer.load(":/images/pic/chara2.png");
    this->imgHome.load(":/images/pic/home.png");
    this->imgSample.load(":/images/pic/capture.png");
    this->setFixedSize(720, 480);
    /* timer for update(redraw) */
    QTimer* timer = new QTimer(this);
    timer->setInterval(16);
    connect(timer, &QTimer::timeout, [this](){
        this->update();
    });
    timer->start();

    this->setFocus();
    this->setFocusPolicy(Qt::StrongFocus);
}

Dialog::~Dialog()
{
}

void Dialog::keyPressEvent(QKeyEvent *event)
{
    if(this->state == State_Main){
        if(event->key() == Qt::Key_Left){
            this->leftPressing = true;
        }
        if(event->key() == Qt::Key_Right){
            this->rightPressing = true;
        }
    } else if(this->state == State_Result){
        if(event->key() == Qt::Key_Enter || event->key() == Qt::Key_Return){
            this->initializeAll();
        } else if(event->key() == Qt::Key_S){
            this->showSample = true;
            this->update();
        }
    } else if(this->state == State_Title){
        if(event->key() == Qt::Key_Left){
            this->leftPressing = true;
            this->state = State_Main;
        }
        if(event->key() == Qt::Key_Right){
            this->rightPressing = true;
            this->state = State_Main;
        }
    }
}

void Dialog::keyReleaseEvent(QKeyEvent *event)
{
    if(this->state == State_Main){
        if(event->key() == Qt::Key_Left){
            this->leftPressing = false;
        }
        if(event->key() == Qt::Key_Right){
            this->rightPressing = false;
        }
    }
}

void Dialog::paintEvent(QPaintEvent *e)
{
    Q_UNUSED(e);
    QPainter p(this);
    p.setRenderHint(QPainter::Antialiasing, true);
    p.fillRect(this->rect(), QColor(36, 36, 36));
    switch(this->state){
    case State_Main:
        this->paintMain(p);
        this->updatePos();
        break;
    case State_Result:
        this->paintResult(p);
        break;
    //case State_Title:
    default:
        this->paintTitle(p);
        break;
    }
}

void Dialog::initializeAll()
{
    this->state = State_Title;
    this->playerPos = ::startPoint;
    this->boundCount = 0;
    this->leftPressing = false;
    this->rightPressing = false;
    /* move up 34 per 40 frames */
    /* 30v + 450a = 34 */
    double t = 40.0;
    this->playerV = ::vel1;
    this->playerA = (34.0 - t * this->playerV) / (t * (t + 1) / 2);
    this->path = QPainterPath();
    this->path.moveTo(this->playerPos);
    this->showSample = false;
    this->dotCount = 0;
    this->dots = std::vector<QPointF>(static_cast<size_t>(::dotMax), this->playerPos);
    this->stars[0].rx() = 540;
    this->stars[0].ry() = 20;
    this->stars[1].rx() = 590;
    this->stars[1].ry() = 81;
    this->stars[2].rx() = 640;
    this->stars[2].ry() = 63;
    this->stars[3].rx() = 690;
    this->stars[3].ry() = 98;
    this->stars[4].rx() = 740;
    this->stars[4].ry() = 50;
    this->stars[5].rx() = 900;
    this->stars[5].ry() = 30;
    this->stars[6].rx() = 1111;
    this->stars[6].ry() = 72;
    for(size_t i = 0; i < static_cast<size_t>(::starMax); i++){
        this->stars[i].rx() += (-6 + qrand() % 12);
        this->stars[i].ry() += (-6 + qrand() % 12);
    }
}

void Dialog::paintTitle(QPainter &p)
{
    /* sky */
    {
        QLinearGradient lg;
        lg.setColorAt(0, QColor(67, 67, 67));
        lg.setColorAt(1, QColor(20, 20, 20));
        lg.setStart(0, 480);
        lg.setFinalStop(0, 0);
        p.fillRect(this->rect(), QBrush(lg));
    }
    /* ground */
    p.setPen(QColor());
    p.fillRect(::groundPos, QColor(176, 110, 53));
    /* stars */
    QPen pen = p.pen();
    pen.setColor(Qt::white);
    pen.setWidth(1);
    p.setPen(pen);
    for(const QPoint& star : this->stars){
        p.drawPoint(QPointF(this->getDrawPosX(star.x()), star.y()));
    }
    /* draw target */
    int i = 1;
    for(QPointF pos : ::targetList){
        p.setBrush(Qt::white);
        p.setPen(Qt::white);
        qreal x = this->getDrawPosX(pos.x());
        p.drawEllipse(QPointF(x, pos.y()), ::radius, ::radius);
        p.drawText(QPoint(static_cast<int>(x), static_cast<int>(pos.y()) + 12), QString("%1").arg(i));
        i++;
    }

    /* player */
    QPointF drawPlayerPos = QPointF(320 - 60, this->playerPos.y() - 60);
    p.drawImage(drawPlayerPos, this->imgPlayer);
    /* title */
    int left = (this->rect().width() - this->imgWayback.width()) / 2;
    p.drawImage(left, 0, this->imgWayback);
    QFont f = p.font();
    f.setPixelSize(12);
    p.setFont(f);
    p.drawText(QRect(0, 300, 640, 75), "Press left or right to start",
               QTextOption(Qt::AlignCenter));
}

void Dialog::paintResult(QPainter &p)
{
    p.fillRect(this->rect(), QColor(67, 67, 67));
    QRect r(0, 20, this->rect().width(), 480 - 100);
    QPen pen;
    pen.setWidth(6.0);
    pen.setColor(Qt::white);
    p.setPen(pen);
    QFont f = p.font();
    f.setPixelSize(32);
    p.setFont(f);
    p.drawPath(this->path);
    p.drawText(r, "WHAT", QTextOption(Qt::AlignTop | Qt::AlignHCenter));
    p.drawText(r, "MEANS TO YOU", QTextOption(Qt::AlignBottom | Qt::AlignHCenter));
    f.setPixelSize(12);
    p.setFont(f);
    if(this->showSample){
        p.drawImage(0, 360, this->imgSample);
    } else {
        p.drawText(this->rect(), "[S]: ???", QTextOption(Qt::AlignBottom | Qt::AlignLeft));
    }
    p.drawText(this->rect(), "[Enter]: Go Title", QTextOption(Qt::AlignBottom | Qt::AlignRight));
}

void Dialog::paintMain(QPainter &p)
{
    /* sky */
    {
        QLinearGradient lg;
        lg.setColorAt(0, QColor(67, 67, 67));
        lg.setColorAt(1, QColor(20, 20, 20));
        lg.setStart(0, 480);
        lg.setFinalStop(0, 0);
        p.fillRect(this->rect(), QBrush(lg));
    }
    /* stars */
    QPen pen = p.pen();
    pen.setColor(Qt::white);
    pen.setWidth(1);
    p.setPen(pen);
    for(const QPoint& star : this->stars){
        p.drawPoint(QPointF(this->getDrawPosX(star.x()), star.y()));
    }
    pen.setColor(Qt::white);
    pen.setWidthF(1.5);
    p.setPen(pen);
    this->path.lineTo(this->playerPos);
    this->dots[static_cast<size_t>(this->dotCount)] = this->playerPos;
    this->dotCount = (this->dotCount + 1) % ::dotMax;
    /* line */
    for(const QPointF& dot : this->dots){
        p.drawPoint(QPointF(this->getDrawPosX(dot.x()), dot.y()));
    }

    /* ground */
    p.setPen(QColor());
    p.fillRect(::groundPos, QColor(176, 110, 53));
    /* draw target */
    int i = 1;
    for(QPointF pos : ::targetList){
        p.setBrush(Qt::white);
        p.setPen(Qt::white);
        qreal x = this->getDrawPosX(pos.x());
        p.drawEllipse(QPointF(x, pos.y()), ::radius, ::radius);
        p.drawText(QPoint(static_cast<int>(x), static_cast<int>(pos.y()) + 12), QString("%1").arg(i));
        i++;
    }

    /* player */
    QPointF drawPlayerPos = QPointF(320 - 60, this->playerPos.y() - 60);
    p.drawImage(drawPlayerPos, this->imgPlayer);
    /* home */
    p.drawImage(QPointF(getDrawPosX(::endPoint.x()), ::endPoint.y() - 72), this->imgHome);
}

void Dialog::updatePos()
{
    if(this->leftPressing){
        this->playerPos.rx() -= ::xspeed;
    }
    if(this->rightPressing){
        this->playerPos.rx() += ::xspeed;
    }
    switch(this->boundCount){
    case 0:/* start to left of "h" */
    {
        if(0 < this->playerV && ::groundPos.top() <= this->playerPos.y() + 60){
            /* from left of "h" to right of "h" */
            /* [about] move up 35 per 30 frames */
            double t = 30.0;
            this->playerV = ::vel1;
            this->playerA = (35.0 - t * this->playerV) / (t * (t + 1) / 2.0);
            this->boundCount = 1;
            this->sndNz1->play();
        }
    }
        break;
    case 1:
    {
        if(0 < this->playerV && ::groundPos.top() <= this->playerPos.y() + 60){
            /* from right of "h" to bottom of o */
            /* [about] move up 35 per 30 frames */
            double t = 60.0;
            this->playerV = -8.0;
            this->playerA = (35.0 - t * this->playerV) / (t * (t + 1) / 2.0);
            this->boundCount = 2;
            this->sndNz1->play();
        }
    }
        break;
    case 2:
    {
        if(0 < this->playerV && ::groundPos.top() <= this->playerPos.y() + 60){
            /* [about] move up 35 per 60 frames */
            double t = 60.0;
            this->playerV = -8.0;//::vel2_slow;
            this->playerA = (35.0 - t * this->playerV) / (t * (t + 1) / 2.0);
            this->boundCount = 3;
            this->sndNz1->play();
        }
    }
        break;
    case 3:
    {
        if(0 < this->playerV && ::groundPos.top() - 48 <= this->playerPos.y() + 60){
            /* [about] move up 21 per 20 frames */
            double t = 20.0;
            this->playerV = ::vel2_slow;
            this->playerA = (21.0 - t * this->playerV) / (t * (t + 1) / 2.0);
            this->boundCount = 4;
            this->sndNz1->play();
        }
    }
        break;
    case 4:
    case 5:
    {
        if(0 < this->playerV && ::groundPos.top() <= this->playerPos.y() + 60){
            /* [about] move up 35 per 35 frames */
            double t = 35.0;
            this->playerV = -10.0;
            this->playerA = (35.0 - t * this->playerV) / (t * (t + 1) / 2.0);
            this->boundCount++;
            this->sndNz1->play();
        }
    }
        break;
    case 6:
    {
        if(0 < this->playerV && ::groundPos.top() <= this->playerPos.y() + 60){
            /* [about] move up 25 per 25 frames */
            double t = 25.0;
            this->playerV = -11.0;
            this->playerA = (25.0 - t * this->playerV) / (t * (t + 1) / 2.0);
            this->boundCount++;
            this->sndNz1->play();
        }
    }
        break;
    case 7:
    {
        if(0 < this->playerV && ::groundPos.top() - 48 <= this->playerPos.y() + 60){
            /* [about] move up 21 per 45 frames */
            double t = 60.0;//45.0;
            this->playerV = -3.5;//::vel2_slow / 2.0;
            this->playerA = (21.0 - t * this->playerV) / (t * (t + 1) / 2.0);
            this->boundCount = 8;
            this->sndNz1->play();
        }
    }
        break;
    case 8:
    {
        if(0 < this->playerV && ::groundPos.top() <= this->playerPos.y() + 60){
            /* [about] move up 35 per 35 frames */
            double t = 35.0;
            this->playerV = -10.0;
            this->playerA = (35.0 - t * this->playerV) / (t * (t + 1) / 2.0);
            this->boundCount++;
            this->sndNz1->play();
        }
    }
        break;
    case 9:
    {
        if(::endPoint.x() < this->playerPos.x()){
            this->boundCount = 10;
            this->playerA = 0.0;
            this->playerV = 0.0;
            this->state = State_Result;
            this->sndDoor->play();
        } else if (0 < this->playerV && ::groundPos.top() <= this->playerPos.y() + 60){
            this->boundCount = 10;
            this->playerA = 0.0;
            this->playerV = 0.0;
            this->state = State_Result;
            this->sndDoor->play();
        }
    }
        break;
    default:
        break;
    }
    this->playerV += this->playerA;
    this->playerPos.ry() += this->playerV;
}

qreal Dialog::getDrawPosX(qreal posX)
{
    qreal diff = this->playerPos.x() - ::startPoint.x();
    qreal x = 320 + posX - ::startPoint.x() - diff;
    return x;
}
